@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
-;; Version: 2.0
+;; Version: 2.3
 ;; Package-Requires: ((emacs "28.1") (compat "30"))
 ;; URL: https://github.com/minad/corfu
 ;; Keywords: abbrev, convenience, matching, completion, text
@@ -344,7 +344,8 @@ settings `corfu-auto-delay', `corfu-auto-prefix' and
     (undecorated . t)
     (cursor-type . nil)
     (no-special-glyphs . t)
-    (desktop-dont-save . t))
+    (desktop-dont-save . t)
+    (inhibit-double-buffering . t)) ;; Avoid display artifacts on X/Gtk builds
   "Default child frame parameters.")
 
 (defvar corfu--buffer-parameters
@@ -366,7 +367,8 @@ settings `corfu-auto-delay', `corfu-auto-prefix' and
     (fringe-indicator-alist (continuation) (truncation))
     (indicate-empty-lines . nil)
     (indicate-buffer-boundaries . nil)
-    (buffer-read-only . t))
+    (buffer-read-only . t)
+    (pixel-scroll-precision-mode . nil))
   "Default child frame buffer parameters.")
 
 (defvar corfu--mouse-ignore-map
@@ -427,7 +429,6 @@ is a prefix length override, which is t for manual completion."
       (cl-pushnew 'corfu-default (alist-get 'default face-remapping-alist))
       buffer)))
 
-(defvar x-gtk-resize-child-frames) ;; Not present on non-gtk builds
 (defvar corfu--gtk-resize-child-frames
   (let ((case-fold-search t))
      ;; XXX HACK to fix resizing on gtk3/gnome taken from posframe.el
@@ -441,6 +442,10 @@ is a prefix length override, which is t for manual completion."
                              (getenv "DESKTOP_SESSION") ""))
          'resize-mode)))
 
+;; Not present on non-gtk/non-x builds
+(defvar x-gtk-resize-child-frames)
+(defvar x-fast-protocol-requests)
+
 ;; Function adapted from posframe.el by tumashu
 (defun corfu--make-frame (frame x y width height)
   "Show current buffer in child frame at X/Y with WIDTH/HEIGHT.
@@ -452,6 +457,7 @@ FRAME is the existing frame."
   (let* ((window-min-height 1)
          (window-min-width 1)
          (inhibit-redisplay t)
+         (x-fast-protocol-requests t)
          (x-gtk-resize-child-frames corfu--gtk-resize-child-frames)
          (before-make-frame-hook)
          (after-make-frame-functions)
@@ -459,6 +465,9 @@ FRAME is the existing frame."
     (unless (and (frame-live-p frame)
                  (eq (frame-parent frame)
                      (and (not (bound-and-true-p exwm--connection)) parent))
+                 ;; Handle mixed tty/graphical sessions
+                 (eq (display-graphic-p frame)
+                     (display-graphic-p parent))
                  ;; If there is more than one window, `frame-root-window' may
                  ;; return nil.  Recreate the frame in this case.
                  (window-live-p (frame-root-window frame)))
@@ -498,7 +507,7 @@ FRAME is the existing frame."
                           unless (equal (alist-get k is) v) collect p)))
       (when diff (modify-frame-parameters frame diff))
       ;; XXX HACK: `set-window-buffer' must be called to force fringe update.
-      (when (or diff (eq (window-buffer win) (current-buffer)))
+      (when (or diff (not (eq (window-buffer win) (current-buffer))))
         (set-window-buffer win (current-buffer)))
       ;; Disallow selection of root window (gh:minad/corfu#63)
       (set-window-parameter win 'no-delete-other-windows t)
@@ -513,7 +522,9 @@ FRAME is the existing frame."
   (make-frame-visible frame)
   ;; Unparent child frame if EXWM is used, otherwise EXWM buffers are drawn on
   ;; top of the Corfu child frame.
-  (when (and (bound-and-true-p exwm--connection) (frame-parent frame))
+  (when (and (bound-and-true-p exwm--connection)
+             (display-graphic-p frame) (frame-parent frame))
+    (redisplay t)
     (set-frame-parameter frame 'parent-frame nil))
   frame)
 
@@ -706,15 +717,9 @@ FRAME is the existing frame."
       ;; Bind non-essential=t to prevent Tramp from opening new connections,
       ;; without the user explicitly requesting it via M-TAB.
       (pcase (let ((non-essential t))
-               ;; XXX Guard against errors during candidate generation.
-               ;; bug#61274: `dabbrev-capf' signals errors.
-               (condition-case err
-                   (if interruptible
-                       (while-no-input (corfu--recompute str pt table pred))
-                     (corfu--recompute str pt table pred))
-                 (error
-                  (message "Corfu completion error: %s" (error-message-string err))
-                  t)))
+               (if interruptible
+                   (while-no-input (corfu--recompute str pt table pred))
+                 (corfu--recompute str pt table pred)))
         ('nil (keyboard-quit))
         ((and state (pred consp))
          (setq corfu--input input)
@@ -777,10 +782,14 @@ FRAME is the existing frame."
   (pcase-let* ((last (min (+ corfu--scroll corfu-count) corfu--total))
                (bar (ceiling (* corfu-count corfu-count) corfu--total))
                (lo (min (- corfu-count bar 1) (floor (* corfu-count corfu--scroll) corfu--total)))
-               (`(,mf . ,acands) (corfu--affixate
-                                  (cl-loop repeat corfu-count
-                                           for c in (nthcdr corfu--scroll corfu--candidates)
-                                           collect (funcall corfu--hilit (substring c)))))
+               (`(,mf . ,acands)
+                (corfu--affixate
+                 (cl-loop
+                  repeat corfu-count for c in (nthcdr corfu--scroll corfu--candidates)
+                  collect (funcall corfu--hilit
+                                   ;; bug#77754: Highlight unquoted string.
+                                   (substring (or (get-text-property
+                                                   0 'completion--unquoted c) c))))))
                (`(,pw ,width ,fcands) (corfu--format-candidates acands))
                ;; Disable the left margin if a margin formatter is active.
                (corfu-left-margin-width (if mf 0 corfu-left-margin-width)))
@@ -852,13 +861,44 @@ the last command must be listed in `corfu-continue-commands'."
   (unless (corfu--range-valid-p)
     (corfu-quit)))
 
+(defun corfu--debug (&rest _)
+  "Debugger used by `corfu--protect'."
+  (let ((inhibit-message t))
+    (require 'backtrace)
+    (declare-function backtrace-to-string "backtrace")
+    (message "Corfu detected an error:\n%s" (backtrace-to-string)))
+  (let (message-log-max)
+    (message "%s %s"
+             (propertize "Corfu detected an error:" 'face 'error)
+             (substitute-command-keys "Press \\[view-echo-area-messages] to see the stack trace")))
+  nil)
+
+(defun corfu--protect (fun)
+  "Protect FUN such that errors are caught.
+If an error occurs, the FUN is retried with `debug-on-error' enabled and
+the stack trace is shown in the *Messages* buffer."
+  (static-if (fboundp 'handler-bind) ;; Available on Emacs 30
+      (ignore-errors
+        (handler-bind ((error #'corfu--debug))
+          (funcall fun)))
+    (when (or debug-on-error (condition-case nil
+                                 (progn (funcall fun) nil)
+                               (error t)))
+      (let ((debug-on-error t)
+            (debugger #'corfu--debug))
+        (condition-case nil
+            (funcall fun)
+          ((debug error) nil))))))
+
 (defun corfu--post-command ()
   "Refresh Corfu after last command."
-  (if (corfu--continue-p)
-      (corfu--exhibit)
-    (corfu-quit))
-  (when corfu-auto
-    (corfu--auto-post-command)))
+  (corfu--protect
+   (lambda ()
+     (if (corfu--continue-p)
+         (corfu--exhibit)
+       (corfu-quit))
+     (when corfu-auto
+       (corfu--auto-post-command)))))
 
 (defun corfu--goto (index)
   "Go to candidate with INDEX."
@@ -886,7 +926,7 @@ See `completion-in-region' for the arguments BEG, END, TABLE, PRED."
   (setq beg (if (markerp beg) beg (copy-marker beg))
         end (if (and (markerp end) (marker-insertion-type end)) end (copy-marker end t))
         completion-in-region--data (list beg end table pred completion-extra-properties))
-  (completion-in-region-mode 1)
+  (completion-in-region-mode)
   (activate-change-group (setq corfu--change-group (prepare-change-group)))
   (setcdr (assq #'completion-in-region-mode minor-mode-overriding-map-alist) corfu-map)
   (add-hook 'pre-command-hook #'corfu--prepare nil 'local)
@@ -989,36 +1029,40 @@ See `completion-in-region' for the arguments BEG, END, TABLE, PRED."
 
 (defun corfu--auto-complete-deferred (&optional tick)
   "Initiate auto completion if TICK did not change."
-  (when (and (not completion-in-region-mode)
-             (or (not tick) (equal tick (corfu--auto-tick))))
-    (pcase (while-no-input ;; Interruptible Capf query
-             (run-hook-wrapped 'completion-at-point-functions #'corfu--capf-wrapper))
-      (`(,fun ,beg ,end ,table . ,plist)
-       (let ((completion-in-region-mode-predicate
-              (lambda ()
-                (when-let ((newbeg (car-safe (funcall fun))))
-                  (= newbeg beg))))
-             (completion-extra-properties plist))
-         (corfu--setup beg end table (plist-get plist :predicate))
-         (corfu--exhibit 'auto))))))
+  (corfu--protect
+   (lambda ()
+     (when (and (not completion-in-region-mode)
+                (or (not tick) (equal tick (corfu--auto-tick))))
+       (pcase (while-no-input ;; Interruptible Capf query
+                (run-hook-wrapped 'completion-at-point-functions #'corfu--capf-wrapper))
+         (`(,fun ,beg ,end ,table . ,plist)
+          (let ((completion-in-region-mode-predicate
+                 (lambda ()
+                   (when-let ((newbeg (car-safe (funcall fun))))
+                     (= newbeg beg))))
+                (completion-extra-properties plist))
+            (corfu--setup beg end table (plist-get plist :predicate))
+            (corfu--exhibit 'auto))))))))
 
 (defun corfu--auto-post-command ()
   "Post command hook which initiates auto completion."
-  (cancel-timer corfu--auto-timer)
-  (when (and (not completion-in-region-mode)
-             (not defining-kbd-macro)
-             (not buffer-read-only)
-             (corfu--match-symbol-p corfu-auto-commands this-command)
-             (corfu--popup-support-p))
-    (if (<= corfu-auto-delay 0)
-        (corfu--auto-complete-deferred)
-      ;; Do not use `timer-set-idle-time' since this leads to
-      ;; unpredictable pauses, in particular with `flyspell-mode'.
-      (timer-set-time corfu--auto-timer
-                      (timer-relative-time nil corfu-auto-delay))
-      (timer-set-function corfu--auto-timer #'corfu--auto-complete-deferred
-                          (list (corfu--auto-tick)))
-      (timer-activate corfu--auto-timer))))
+  (corfu--protect
+   (lambda ()
+     (cancel-timer corfu--auto-timer)
+     (when (and (not completion-in-region-mode)
+                (not defining-kbd-macro)
+                (not buffer-read-only)
+                (corfu--match-symbol-p corfu-auto-commands this-command)
+                (corfu--popup-support-p))
+       (if (<= corfu-auto-delay 0)
+           (corfu--auto-complete-deferred)
+         ;; Do not use `timer-set-idle-time' since this leads to
+         ;; unpredictable pauses, in particular with `flyspell-mode'.
+         (timer-set-time corfu--auto-timer
+                         (timer-relative-time nil corfu-auto-delay))
+         (timer-set-function corfu--auto-timer #'corfu--auto-complete-deferred
+                             (list (corfu--auto-tick)))
+         (timer-activate corfu--auto-timer))))))
 
 (defun corfu--auto-tick ()
   "Return the current tick/status of the buffer.
@@ -1030,7 +1074,7 @@ Auto completion is only performed if the tick did not change."
 WIDTH is the width of the popup.
 The current candidate CURR is highlighted.
 A scroll bar is displayed from LO to LO+BAR."
-  (let ((lh (default-line-height)))
+  (let ((lh (max (default-line-height) (cdr (posn-object-width-height pos)))))
     (with-current-buffer (corfu--make-buffer " *corfu*")
       (let* ((ch (default-line-height))
              (cw (default-font-width))
@@ -1431,7 +1475,7 @@ local `completion-at-point-functions'."
                                   ('nil 0)
                                   ((pred symbolp) (and (derived-mode-p p) t))
                                   (`(not . ,m) (and (seq-some #'derived-mode-p m) 0)))))))
-    (corfu-mode 1)))
+    (corfu-mode)))
 
 (defun corfu--minibuffer-on ()
   "Enable `corfu-mode' in the minibuffer respecting `global-corfu-minibuffer'."
@@ -1439,7 +1483,7 @@ local `completion-at-point-functions'."
              (if (functionp global-corfu-minibuffer)
                  (funcall global-corfu-minibuffer)
                (local-variable-p 'completion-at-point-functions)))
-    (corfu-mode 1)))
+    (corfu-mode)))
 
 ;; Do not show Corfu commands with M-X
 (dolist (sym '( corfu-next corfu-previous corfu-first corfu-last corfu-quit corfu-reset
